/******************************************************************************
  Skeleton Controller

  Monitors a motion sensor.
  Upon motion detection:
    Activate opening sounds.
    Activate the motor to open the coffin.
  Once the sensors detect that the coffin is open.
    Stop the motor.
    Activate skeleton sounds.
    Activate the skeleton eyes.
    Wait 9 seconds.
  Deactivate the skeleton eyes.
  Activate closing sounds.
  Activate the motor to close the coffin.
  Once the sensors detect that the coffin is closed. Stop the motor.

  Timeouts:
  If the motor runs for over 6 seconds, this is too long. Something is wrong.
  Stop the motor and enter a MOTOR_TIMEOUT state.
 ****************************************************************************/
#include "SkelEnum.h"  //enumeration of the different states available

//PINS
//pins 0, 1 are TX/RX
//nano pins A0-A5 can be used as digital pins

//sensors
#define S_MOTION 3
#define S_LID_OPEN_SONIC 4  //input from the ultrasonic sensor
#define S_LID_OPEN_SWITCH 5 //input from the lid-open switch
#define S_LID_CLOSED 6

//output
#define O_EYES 2

//1 pin could trigger multiple 5mA relays, not doing that though
#define O_LID_MOTOR_A 7
#define O_LID_MOTOR_B 8  //This is the COMMON relay. It is the ground/negative that switches to either side of the motor.
#define O_LID_MOTOR_C 9
#define O_TRIGGER_SONIC 10  //Triggers the ultrasonic sensor to take a distance measurement

//sound outputs
//Use the analog pins to keep the sound outputs all on one side of the Arduino.
#define O_SOUND_OPENING A0
#define O_SOUND_SKELETON A1
#define O_SOUND_CLOSING A2
#define O_SOUND_AMBIANCE0 A3
#define O_SOUND_AMBIANCE1 A4
#define O_SOUND_AMBIANCE2 A5

//Holds the current state
StateEnum currentState;

int LID_MOTOR_OPEN_TIMEOUT = 6000; //ms, max time the lid can take to open
int LID_MOTOR_CLOSE_TIMEOUT = 6000; //ms, max time the lid can take to close

//Timing, all in milliseconds
int COFFIN_IDLE_TIME = 5000; //time when the coffin starts, before it does anything
int COFFIN_OPEN_TIME = 9000; //time the lid will remain open for, 9 seconds
int COFFIN_MIN_CLOSE_TIME = 15000; //min time the coffin will stay closed between openings
int COFFIN_WAIT_BEFORE_AMBIANCE = 15000; //time after which the coffin will play ambiance sounds
int DURING_AFTER_AMBIANCE_TIME = 10000; //time during and after ambiance
int AUDIO_TRIGGER_TIME = 50; //time to hold the pin low in order to trigger the audio
int LID_SOUND_DELAY_TIME = 200; //delay after triggering the audio, before activating the motor

//Ultrasonic sensor constants
//int (2 bytes) range: -32,768 to 32,767
float TEMPERATURE_F = 60;
float TEMPERATURE_C = (TEMPERATURE_F - 32.0) * (5.0 / 9.0);
//Precalculate the speed of sound given the temperature. Add .5 to round when truncating the decimal off.
float SPEED_OF_SOUND_M = (331.3 + (0.606 * TEMPERATURE_C));
float SPEED_OF_SOUND_MM_FLOAT = SPEED_OF_SOUND_M * 1000.0;  //in millimeters
unsigned long SPEED_OF_SOUND_MM = (unsigned long)(SPEED_OF_SOUND_MM_FLOAT + 0.5);  //in millimeters
unsigned long DISTANCE_TO_STEEL_BAR_MM = 40;  //in millimeters

//for the morse code
int DOT = 130;
int DASH = DOT*3;

//holds the current start of the coffin
unsigned long currentStateStartedAt;

/******************************************************************************
 * The setup function runs once when you press reset or power the board
 *****************************************************************************/
void setup() {
  Serial.begin(9600); //this starts the serial monitor

  //input
  pinMode(S_MOTION, INPUT);  //high = active
  pinMode(S_LID_OPEN_SONIC, INPUT);  //high = active
  pinMode(S_LID_OPEN_SWITCH, INPUT_PULLUP);  //low = active
  pinMode(S_LID_CLOSED, INPUT_PULLUP); //low = active

  //output
  pinMode(LED_BUILTIN, OUTPUT);  //pin 13
  digitalWrite(LED_BUILTIN, LOW);

  pinMode(O_LID_MOTOR_A, OUTPUT);  //high = active
  pinMode(O_LID_MOTOR_B, OUTPUT);
  pinMode(O_LID_MOTOR_C, OUTPUT); //high = active
  pinMode(O_TRIGGER_SONIC, OUTPUT);

  digitalWrite(O_LID_MOTOR_A, LOW);
  digitalWrite(O_LID_MOTOR_B, LOW);
  digitalWrite(O_LID_MOTOR_C, LOW);
  digitalWrite(O_TRIGGER_SONIC, LOW);

  pinMode(O_SOUND_OPENING, OUTPUT);  //low = active
  pinMode(O_SOUND_SKELETON, OUTPUT);  //low = active
  pinMode(O_SOUND_CLOSING, OUTPUT);  //low = active
  pinMode(O_SOUND_AMBIANCE0, OUTPUT);  //low = active
  pinMode(O_SOUND_AMBIANCE1, OUTPUT);  //low = active
  pinMode(O_SOUND_AMBIANCE2, OUTPUT);  //low = active
  digitalWrite(O_SOUND_OPENING, HIGH);
  digitalWrite(O_SOUND_SKELETON, HIGH);
  digitalWrite(O_SOUND_CLOSING, HIGH);
  digitalWrite(O_SOUND_AMBIANCE0, HIGH);
  digitalWrite(O_SOUND_AMBIANCE1, HIGH);
  digitalWrite(O_SOUND_AMBIANCE2, HIGH);

  pinMode(O_EYES, OUTPUT);  //high = active
  digitalWrite(O_EYES, LOW);

  //set the coffin state to idle
  SetState(ST_IDLE);

  //Serial.println("setup: SPEED_OF_SOUND_MM: " + String(SPEED_OF_SOUND_MM));
  Serial.println("setup: Complete");
}//setup

/******************************************************************************
 * The loop function runs over and over again forever.
 * When motion is detected, open the lid, start the sounds, start the eyes.
 *****************************************************************************/
void loop() {
  //State Machine
  //Tried using a switch statement, encountered strange issues with states not
  //being entered when they should have. Changed to 'if' statements.

  //IDLE: Just started
  if (currentState == ST_IDLE) {
    if(IsTimeoutReached(COFFIN_IDLE_TIME)) {  //wait for the idle time before doing stuff
      if (IsLidClosed()) {
        SetState(ST_CLOSED_WAITING);
      }
      if (IsLidOpen()) {
        SetState(ST_OPEN);
      }
    }
  }

  //CLOSED_WAITING: Coffin is closed and waiting for motion.
  //If no motion for 15 seconds, play some ambiance.
  else if (currentState == ST_CLOSED_WAITING) {
    if (IsMotion()) {  //any non-zero number evaluates to TRUE, HIGH evaluates to true
      OpenCoffin();
      SetState(ST_OPENING);
    }
    else { //if no motion for 15 seconds, play some ambiance
      if(IsTimeoutReached(COFFIN_WAIT_BEFORE_AMBIANCE)) {
        SetState(ST_PLAY_AMBIANCE);
      }
    }
  }

  //PLAY_AMBIANCE: Trigger ambiance sounds.
  else if (currentState == ST_PLAY_AMBIANCE) {
    PlayAmbianceSound();
    SetState(ST_AFTER_AMBIANCE);
  }

  //AFTER_AMBIANCE: Coffin is closed and waiting for motion. Ambiance sounds were triggered.
  //Wait for them to play, and wait a few more seconds. Then return to ST_CLOSED_WAITING.
  //If motion is detected, go to ST_OPENING.
  else if (currentState == ST_AFTER_AMBIANCE) {
    if (IsMotion()) {
      OpenCoffin();
      SetState(ST_OPENING);
    }
    else { //if no motion
      if(IsTimeoutReached(DURING_AFTER_AMBIANCE_TIME)) {  //stay in this state for 10 seconds,
        SetState(ST_CLOSED_WAITING);
      }
    }
  }

  //OPENING: Coffin lid is in the process of opening.
  else if (currentState == ST_OPENING) {
    if(IsTimeoutReached(LID_MOTOR_OPEN_TIMEOUT)) {
      LidMotorStop();
      SetState(ST_MOTOR_TIMEOUT);
      return;
    }

    if (IsLidOpen()) {
      LidMotorStop();
      PlaySound(O_SOUND_SKELETON);
      TurnOnEyes(true);
      SetState(ST_OPEN);
    }
  }

  //OPEN: Wait some seconds with the lid open.
  else if (currentState == ST_OPEN) {
    if (IsTimeoutReached(COFFIN_OPEN_TIME)) {
      TurnOnEyes(false);
      PlaySound(O_SOUND_CLOSING);
      delay(LID_SOUND_DELAY_TIME);  //give the sounds a chance to start
      LidMotorClose();
      SetState(ST_CLOSING);
    }
  }

  //CLOSING: The lid is in the process of closing.
  else if (currentState == ST_CLOSING) {
    if(IsTimeoutReached(LID_MOTOR_CLOSE_TIMEOUT)) {
      LidMotorStop();
      SetState(ST_MOTOR_TIMEOUT);
      return;
    }

    if (IsLidClosed()) {
      LidMotorStop();
      SetState(ST_JUST_CLOSED);
    }
  }

  //JUST_CLOSED: wait for a little bit before opening again.
  else if (currentState == ST_JUST_CLOSED) {
    if (IsTimeoutReached(COFFIN_MIN_CLOSE_TIME)) {
      SetState(ST_CLOSED_WAITING);
    }
  }

  //MOTOR_TIMEOUT: The motor was running, but the open or closed sensors did not register before a timeout.
  //Remain in this state until reset.
  else if (currentState == ST_MOTOR_TIMEOUT) {
    Serial.println(StateEnumStringArray[currentState]);
    BlinkSOS(LED_BUILTIN);
  }

  else {
    Serial.println("In the default");
    BlinkSOS(LED_BUILTIN);
  } //switch
}//loop

/******************************************************************************
 * Set the state.
 * Set the start time for this state.
 *****************************************************************************/
void SetState(StateEnum newState)
{
  //Serial.print("SetState: ");
  //Serial.print(StateEnumStringArray[currentState]);
  //Serial.print(" => ");
  //Serial.println(StateEnumStringArray[newState]);

  currentStateStartedAt = millis();
  currentState = newState;
}

/******************************************************************************
 * Check if a specific amount of time has passed since the coffin entered
 * this state.
 * Checks the current time vs. the 'currentStateStartedAt' global variable.
 * Return true if the timeout amount of time has passed.
 *
 * int timeout: This is the amount of time to check for in milliseconds.
 ******************************************************************************/
bool IsTimeoutReached(int timeout)
{
  int timeRunning = millis() - currentStateStartedAt;

  //Serial.println("IsTimeoutReached(): elapsed: " + String(timeRunning) + " of " + String(timeout));
  if (timeRunning >= timeout) {
    return true;
  }
  else {
    return false;
  }
}

/******************************************************************************
 * If motion is detected, open the coffin.
 *
 * Opening the coffin entails playing sounds and opening the lid.
 *****************************************************************************/
void OpenCoffin()
{
  PlaySound(O_SOUND_OPENING);
  delay(LID_SOUND_DELAY_TIME);  //give the sounds a chance to start
  LidMotorOpen();
}

/******************************************************************************
 * Run the motor to open the lid.
 *
 * Connect common to the right side of the motor.
 * Connect +12V to the left.
 *****************************************************************************/
void LidMotorOpen()
{
  //Serial.println("LidMotorOpen()");
  digitalWrite(O_LID_MOTOR_C, LOW);  //double checking this is off

  digitalWrite(O_LID_MOTOR_B, LOW); //common, connected to normally open (NO)
  digitalWrite(O_LID_MOTOR_A, HIGH); //+12V
}

/******************************************************************************
 * Run the motor to close the lid.
 *
 * Connect common to the left side of the motor.
 * Connect +12V to the right.
 *****************************************************************************/
void LidMotorClose()
{
  //Serial.println("LidMotorClose()");
  digitalWrite(O_LID_MOTOR_A, LOW);  //double checking this is off

  digitalWrite(O_LID_MOTOR_B, HIGH); //common, connected to normally closed (NC)
  digitalWrite(O_LID_MOTOR_C, HIGH); //+12V
}

/******************************************************************************
 * Stop the motor.
 *
 * First disconnect both +12V.
 * Then set common (relay B) to the LOW state.
 *****************************************************************************/
void LidMotorStop()
{
  //Serial.println("LidMotorStop()");

  //disconnect both +12V sources
  digitalWrite(O_LID_MOTOR_A, LOW);
  digitalWrite(O_LID_MOTOR_C, LOW);

  //Whatever setting comm (relay B) is on, let it stay there for a microsecond for the snubber to work.
  //Also don't want a short circuit (if C was energized and comm (relay B) connected to it).
  //Relay release time is under 5ms.
  delay(10);
  digitalWrite(O_LID_MOTOR_B, LOW);
}

/******************************************************************************
 * Check if the coffin is closed.
 * Returns true if open.
 *
 *****************************************************************************/
bool IsLidClosed()
{
  //Serial.println("IsLidClosed()");
  return !digitalRead(S_LID_CLOSED); //if sensor returns low, lid is closed
}

/******************************************************************************
 * Checks if the coffin is fully open by checking if the steel bar is in the
 * 'open coffin' position. The steel bar is in the open position if it has
 * triggered the S_LID_OPEN_SWITCH or if the ultrasonic sensor reports
 * that the steel bar is near to it.
 * 
 * Returns true if open.
 *
 *****************************************************************************/
bool IsLidOpen()
{
  //Serial.println("IsLidOpen()");
  return (!digitalRead(S_LID_OPEN_SWITCH) ||    //if switch returns low, lid is fully open
          IsSonicProximity(O_TRIGGER_SONIC, S_LID_OPEN_SONIC, DISTANCE_TO_STEEL_BAR_MM) ||
          !digitalRead(S_LID_OPEN_SWITCH)); //check again because there is a small delay in the proximity check
}

/******************************************************************************
 * Check if motion has been detected.
 * Returns true if motion detected.
 *
 * sensor HIGH = active
 *****************************************************************************/
bool IsMotion()
{
  return digitalRead(S_MOTION);
}

/******************************************************************************
 * Turn the skeleton eyes on or off.
 *
 * True means turn on the eyes.
 * output HIGH = turn on.
 *****************************************************************************/
void TurnOnEyes(bool input)
{
  //Serial.println("TurnOnEyes(): " + String(input));
  if(input) {
    digitalWrite(O_EYES, HIGH);
  }
  else {
    digitalWrite(O_EYES, LOW);
  }
}

/******************************************************************************
 * Play 1 of 3 ambiance sounds. Select randomly.
 *****************************************************************************/
//A3 is a #define in Arduino library
void PlayAmbianceSound()
{
  //Serial.println("PlayAmbianceSound()");
  long rand = random(3);
  switch (rand) {
    case 0:
      PlaySound(O_SOUND_AMBIANCE0);
      break;
    case 1:
      PlaySound(O_SOUND_AMBIANCE1);
      break;
    case 2:
      PlaySound(O_SOUND_AMBIANCE2);
      break;
  }
}

/******************************************************************************
 * Play a sound once.
 *
 * The sound chip is set to play a sound all the way through once started
 * (unless another sound is triggered).
 * So turn the sound on, have a brief delay, then turn it off.
 * This method delays for 50ms.
 *
 * int soundPin: The Arduino pin to output the signal on.
 *
 * LOW = play sound
 *****************************************************************************/
void PlaySound(int soundPin)
{
  //Serial.println("PlaySound(): " + String(soundPin));
  digitalWrite(soundPin, LOW);
  delay(AUDIO_TRIGGER_TIME);
  digitalWrite(soundPin, HIGH);
}

/******************************************************************************
 * Check if the steel bar is within a certain distance of the ultrasonic sensor.
 * 
 * The untrasonic sensor has 4 pins/connections. 
 *  VCC (Power), Trig (Trigger), Echo (Receive), and GND (Ground)
 *  Trigger: Triggers the ultrasonic sensor to take a reading. 
 *    To trigger, the Arduino puts a HIGH (+5V) voltage on this pin.
 *  Echo: The ultrasonic sensor holds this HIGH for the same length of time
 *    that the sound burst took to go from the ultrasonic sensor to the object 
 *    and back to the ultrasonic sensor ('flight time').
 * 
 * To measure distance with an ultrasonic sensor
 * 1) Set trigger pin LOW for at least 2-4 microseconds.
 * 2) Set trigger pin HIGH for 10 microseconds to trigger the ultrasonic sensor to send a burst of sound.
 * 3) Use pulseIn() to wait for, and then measure, how long the ultrasonic sensor holds the echo pin HIGH.
 * 4) If pulseIn() returns a 0, the echo pin was never set to HIGH by the ultrasonic sensor. This means
 *    a) there was no object that reflected the sound burst sent out by the ultrasonic sensor, or
 *    b) any objects were far away. They did not reflect an echo before the timeout was reached.
 *       PulseIn() has a default timeout of 1 second. The ultrasonic sensor's max range is 4 meters.
 * 5) Given the duration that the ultrasonic sensor held the echo pin HIGH (as returned by PulseIn()),
 *      calculate the distance that sound would have travelled over that time period.
 *
 * int triggerPin: Pin to trigger the ultrasonic sensor with
 * int readPin: Pin to read the echo time from
 * int proximityDistanceMM: Distance to check for, in the same units as the speed of sound
 *****************************************************************************/
bool IsSonicProximity(int triggerPin, int readPin, unsigned long proximityDistanceMM)
{
  //I attempted to calculate the timeout based on the maximum distance I was looking for. 
  //I.E. If no echo was received in the amount of time it would have taken sound to cover 
  //the max distance, then the steel bar must not be close, so timeout.
  //However, this timeout was too short. This technique did not work. 
  //The timeout always occured before a signal was received (timeout too short).
  //unsigned long pulseTimeoutInMicroseconds = GetSoundRoundTripTime(SPEED_OF_SOUND_MM, proximityDistanceMM) + PULSEIN_ADDITIONAL_TIMEOUT;

  //The timeout needs to be longer than one might expect.
  //4000 timeout seems to correspond to a 750us (microseconds) echo duration
  //3700, 450us
  //3600, 338us
  //So about 3300us is required as a baseline timeout to wait
  //for the sensor to begin sending back an echo signal.
  //Perhaps this is just the time that the sensor takes to operate.

  //The timeout is the maximum amount of time we will wait for an echo from the steel bar.
  //If the timeout passes and no echo, then it must be far enough away that the coffin is not open.
  unsigned long pulseTimeoutInMicroseconds = 3700L;  //3.7ms
  //Serial.println("IsSonicProximity(): pulseTimeoutInMicroseconds: "+ String(pulseTimeoutInMicroseconds));

  //low before triggering ultrasonic sensor
  digitalWrite(triggerPin, LOW);
  delayMicroseconds(4);

  //trigger the ultrasonic sensor
  digitalWrite(triggerPin, HIGH);
  delayMicroseconds(10);  //SONIC_TRIGGER_DELAY = 10;  //in microseconds
  digitalWrite(triggerPin, LOW);

  //DO NOT PRINT OUT HERE!!!!!!!!!
  //Printing takes about 400~600 microseconds and seems to disrupt reading the echo pin
  //using PulseIn().

  //Read the length of time to get an echo.
  //PulseIn: Works for pulses 10 microseconds to 3 minutes in length, default timeout is one second.
  unsigned long echoDurationInMicroseconds = pulseIn(readPin, HIGH, pulseTimeoutInMicroseconds);
  if (echoDurationInMicroseconds == 0)  //no echo, or timeout occured
    return false;

  //calculate the distance given the echo time
  unsigned long distanceToObjectMM = GetSoundDistance(SPEED_OF_SOUND_MM, echoDurationInMicroseconds);
  if (distanceToObjectMM <=  proximityDistanceMM)
    return true;
  else
    return false;
}

/******************************************************************************
 * Calculate the time it would take sound to travel a round trip distance.
 * Speed of sound and the distance should be in the same units.
 * Returns: the time, in microseconds, that sound should take to cover the round trip distance
 *
 * unsigned long speedOfSoundInMillimeters: speed of sounds in millimeters per second
 * unsigned long distanceInMillimeters: distance in millimeters
 *****************************************************************************/
unsigned long GetSoundRoundTripTime(unsigned long speedOfSoundInMillimeters, unsigned long distanceInMillimeters)
{
  unsigned long timeInMicroseconds = (distanceInMillimeters * 2 * 1000000) / speedOfSoundInMillimeters;
  return timeInMicroseconds;
}

/******************************************************************************
 * Calculate the distance to an object, given the echo duration 
 * 
 * Echo duration: Time sound takes to go to the object, bounce off, and go back to the sensor.
 * Distance = (echo duration × velocity of sound) / 2
 * Returns: the distance in the same units as the speed of sound.
 * 
 * unsigned long speedOfSoundInMillimeters: speed of sounds in millimeters per second
 * unsigned long echoDurationInMicroseconds: Time, in microseconds, it took for sound to make a round trip to the object.
 *****************************************************************************/
unsigned long GetSoundDistance(unsigned long speedOfSoundInMillimeters, unsigned long echoDurationInMicroseconds)
{
  unsigned long distanceInMillimeters = ((speedOfSoundInMillimeters * echoDurationInMicroseconds) / 2) / 1000000;
  return distanceInMillimeters;
}

/******************************************************************************
 * Blink SOS
 *
 * The length of a dot is 1 time unit.
 * A dash is 3 time units.
 * The space between symbols (dots and dashes) of the same letter is 1 time unit.
 * The space between letters is 3 time units.
 * The space between words is 6 time units.
 * https://www.infoplease.com/encyclopedia/science/engineering/electrical/morse-code
 * https://www.codebug.org.uk/learn/step/541/morse-code-timing-rules/
 *****************************************************************************/
void BlinkSOS(int pin) {
  flash3(pin, DOT);  //S
  flash3(pin, DASH); //O
  flash3(pin, DOT);  //S
  delay(DASH); //space between words is 6 time units. There was a dash at the end of the letter, so add one more.
}

/******************************************************************************
 * Flash an 'S' or and 'O'.
 *
 * These are each 3 flashes.
 * S = ...
 * O = ---
 *
 * int pin: pin to output on
 * int dotOrDash: send 3 dots or dashes
 *****************************************************************************/
void flash3(int pin, int dotOrDash) {
  for (int i=0;i<3;i++) {
    digitalWrite(pin,HIGH);
    delay(dotOrDash);
    digitalWrite(pin,LOW);
    delay(DOT);  //space between symbols (dots and dashes) of the same letter is 1 time unit.
  }
  delay(DASH-DOT);  //between letters is 3 time units. We just waited a dot, so subtract that.
}
